import io
import json
from http import HTTPStatus

import pandas as pd
import requests

from enfobench.core import ModelInfo
from enfobench.evaluation.server import EnvironmentInfo


def df_to_buffer(df: pd.DataFrame) -> io.BytesIO:
    buffer = io.BytesIO()
    df.to_parquet(buffer, index=True)
    buffer.seek(0)
    return buffer


def metadata_to_buffer(metadata: dict) -> io.BytesIO:
    buffer = io.BytesIO()
    buffer.write(json.dumps(metadata).encode())
    buffer.seek(0)
    return buffer


class ForecastClient:
    def __init__(self, host: str = "localhost", port: int = 3000, *, use_https: bool = False, client=None):
        self.base_url = f"{'https' if use_https else 'http'}://{host}:{port}"
        self._session = requests.Session() if client is None else client

    def forecast(
        self,
        horizon: int,
        history: pd.DataFrame,
        past_covariates: pd.DataFrame | None = None,
        future_covariates: pd.DataFrame | None = None,
        metadata: dict | None = None,
        level: list[int] | None = None,
    ) -> pd.DataFrame:
        params: dict[str, int | list[int]] = {
            "horizon": horizon,
        }
        if level is not None:
            params["level"] = level

        files = {
            "history": df_to_buffer(history),
        }
        if past_covariates is not None:
            files["past_covariates"] = df_to_buffer(past_covariates)
        if future_covariates is not None:
            files["future_covariates"] = df_to_buffer(future_covariates)
        if metadata is not None:
            files["metadata"] = metadata_to_buffer(metadata)

        response = self._session.post(
            url=f"{self.base_url}/forecast",
            params=params,
            files=files,
        )
        if response.status_code != HTTPStatus.OK:
            response.raise_for_status()

        df = pd.DataFrame.from_records(response.json()["forecast"])
        df["timestamp"] = pd.to_datetime(df["timestamp"])
        df.set_index("timestamp", inplace=True)
        return df

    def info(self) -> ModelInfo:
        response = self._session.get(f"{self.base_url}/info")
        if response.status_code != HTTPStatus.OK:
            response.raise_for_status()

        return ModelInfo(**response.json())

    def environment(self) -> EnvironmentInfo:
        response = self._session.get(f"{self.base_url}/environment")
        if response.status_code != HTTPStatus.OK:
            response.raise_for_status()

        return EnvironmentInfo(**response.json())



import warnings

import pandas as pd


def create_perfect_forecasts_from_covariates(
    past_covariates: pd.DataFrame,
    *,
    horizon: pd.Timedelta,
    step: pd.Timedelta,
    start: pd.Timestamp | None = None,
) -> pd.DataFrame:
    """Create forecasts from covariates.

    Sometimes external forecasts are not available for the entire horizon. This function creates
    external forecast dataframe from external covariates as a perfect forecast.

    Args:
        past_covariates: The external covariates.
        horizon: The forecast horizon.
        step: The step size between forecasts.
        start: The start date of the forecast. If None, the first date of the covariates is used.

    Returns:
        The external forecast dataframe.
    """
    start = start or past_covariates.index[0]

    if start < past_covariates.index[0]:
        msg = f"start={start} must be after the start of the past_covariates={past_covariates.index[0]}"
        raise ValueError(msg)

    if start > past_covariates.index[-1]:
        msg = f"start={start} must be before the end of the past_covariates={past_covariates.index[-1]}"
        raise ValueError(msg)

    last_date = past_covariates.index[-1]

    forecasts = []
    while start + horizon <= last_date:
        forecast = past_covariates.loc[start : start + horizon]
        forecast.rename_axis("timestamp", inplace=True)
        forecast.reset_index(inplace=True)
        forecast.insert(0, "cutoff_date", pd.to_datetime(start, unit="ns"))
        forecast.set_index(["cutoff_date", "timestamp"], inplace=True)

        if len(forecast) == 0:
            warnings.warn(
                f"Covariates not found for {start} - {start + horizon}, cannot make forecast at step {start}",
                UserWarning,
                stacklevel=2,
            )
        else:
            forecasts.append(forecast)
        start += step

    forecast_df = pd.concat(forecasts, ignore_index=False)
    forecast_df.reset_index(inplace=True)
    forecast_df.sort_values(by=["cutoff_date", "timestamp"], inplace=True)
    return forecast_df



import logging
import warnings
from abc import ABCMeta
from pathlib import Path

import pandas as pd
from pandas.tseries.frequencies import to_offset

logger = logging.getLogger(__name__)


class Subset(metaclass=ABCMeta):  # noqa: B024
    """Subset class representing one of the subset of the HuggingFace dataset.

    Args:
        file_path: The path to the subset file.
    """

    def __init__(self, file_path: Path | str) -> None:
        file_path = Path(file_path).resolve()
        if not file_path.is_file() or not file_path.exists():
            msg = "Please provide the existing file where the subset is located."
            raise ValueError(msg)
        self.file_path = file_path

    def __repr__(self):
        return f"{self.__class__.__name__}(file_path={self.file_path})"

    def read(self) -> pd.DataFrame:
        """Read the subset from the file."""
        return pd.read_parquet(self.file_path)


class Dataset:
    """Dataset class representing a collection of data required for forecasting task.

    Args:
        target: The target variable.
        past_covariates: The past covariates.
        future_covariates: The future covariates.
        metadata: The metadata.
    """

    def __init__(
        self,
        target: pd.DataFrame,
        past_covariates: pd.DataFrame | None = None,
        future_covariates: pd.DataFrame | None = None,
        metadata: dict | None = None,
    ):
        self._target = self._check_target(target.copy())
        self._first_available_target_date: pd.Timestamp = self._target.index[0]
        self._last_available_target_date: pd.Timestamp = self._target.index[-1]

        self._past_covariates = (
            self._check_past_covariates(past_covariates.copy()) if past_covariates is not None else None
        )

        self._future_covariates = (
            self._check_external_forecasts(future_covariates.copy()) if future_covariates is not None else None
        )
        self.metadata = metadata

    @property
    def target_available_since(self) -> pd.Timestamp:
        """Returns the first available target date."""
        return self._first_available_target_date

    @property
    def target_available_until(self) -> pd.Timestamp:
        """Returns the last available target date."""
        return self._last_available_target_date

    @property
    def target_freq(self) -> str:
        """Returns the frequency of the target."""
        return self._target.index.inferred_freq

    @staticmethod
    def _check_target(y: pd.DataFrame) -> pd.DataFrame:
        if isinstance(y, pd.Series):
            msg = "Target is a Series, converting to DataFrame."
            warnings.warn(msg, UserWarning, stacklevel=2)
            y = y.to_frame("y")

        if not isinstance(y.index, pd.DatetimeIndex):
            msg = f"Target dataframe must have DatetimeIndex, have {y.index.__class__.__name__}."
            raise ValueError(msg)

        if y.index.inferred_freq is None:
            msg = "Frequency cannot be inferred from target's index, resampling target."
            warnings.warn(msg, UserWarning, stacklevel=2)
            freq = to_offset(y.index[1] - y.index[0]).freqstr
            freq = "1" + freq if not freq[0].isdigit() else freq
            y = y.resample(freq).asfreq()

        y.rename_axis("timestamp", inplace=True)
        y.sort_index(inplace=True)
        return y

    def _check_past_covariates(self, X: pd.DataFrame) -> pd.DataFrame:  # noqa: N803
        if not isinstance(X.index, pd.DatetimeIndex):
            msg = f"Past covariates must have DatetimeIndex, have {X.index.__class__.__name__}."
            raise ValueError(msg)

        X.rename_axis("timestamp", inplace=True)
        X.sort_index(inplace=True)

        if X.index[0] > self._first_available_target_date:
            msg = "Covariates should be provided for the full target timeframe, covariates start after target values."
            warnings.warn(msg, UserWarning, stacklevel=2)

        if X.index[-1] < self._last_available_target_date:
            msg = "Covariates should be provided for the full target timeframe, covariates end before target values."
            warnings.warn(msg, UserWarning, stacklevel=2)
        return X

    def _check_external_forecasts(self, X: pd.DataFrame) -> pd.DataFrame:  # noqa: N803
        first_forecast_date = X.cutoff_date.min()
        last_forecast_date = X.cutoff_date.max()
        last_forecast_end_date = X[X.cutoff_date == last_forecast_date].timestamp.max()

        if first_forecast_date > self._first_available_target_date:
            msg = (
                "External forecasts should be provided for the full target timeframe, "
                "forecasts start after target values."
            )
            warnings.warn(msg, UserWarning, stacklevel=2)

        if last_forecast_end_date < self._last_available_target_date:
            msg = (
                "External forecasts should be provided for the full target timeframe, "
                "forecasts end before target values."
            )
            warnings.warn(msg, UserWarning, stacklevel=2)

        return X

    def _check_cutoff_in_rage(self, cutoff_date: pd.Timestamp):
        if cutoff_date < self._first_available_target_date:
            msg = f"Cutoff date is before the start date: {cutoff_date} < {self._first_available_target_date}."
            raise IndexError(msg)

        if cutoff_date > self._last_available_target_date:
            msg = f"Cutoff date is after the end date: {cutoff_date} > {self._last_available_target_date}."
            raise IndexError(msg)

    def get_history(self, cutoff_date: pd.Timestamp) -> pd.DataFrame:
        """Returns the history of the target variable up to the cutoff date.

        The cutoff date is the timestamp when the forecast is made,
        therefore the cutoff_date is not included in the history.

        Args:
            cutoff_date: The cutoff date.

        Returns:
            The history of the target variable up to the cutoff date.
        """
        self._check_cutoff_in_rage(cutoff_date)
        return self._target[self._target.index < cutoff_date]

    def get_past_covariates(self, cutoff_date: pd.Timestamp) -> pd.DataFrame | None:
        """Returns the past covariates for the cutoff date.

        The cutoff date is the timestamp when the forecast is made.
        As the covariates are weather parameters measured at the indicated timestamp,
        the cutoff_date is included in the past covariates.

        Args:
            cutoff_date: The cutoff date.

        Returns:
            The past covariates up until the cutoff date.
        """
        if self._past_covariates is None:
            return None

        self._check_cutoff_in_rage(cutoff_date)
        return self._past_covariates[self._past_covariates.index <= cutoff_date]

    def get_future_covariates(self, cutoff_date: pd.Timestamp) -> pd.DataFrame | None:
        """Returns the future covariates for the cutoff date.

        The cutoff date is the timestamp when the forecast is made.
        As the covariates are weather parameters measured at the indicated timestamp,
        the cutoff_date is not included in the future covariates.

        Args:
            cutoff_date: The cutoff date.

        Returns:
            The last external forecasts made before the cutoff date.
        """
        if self._future_covariates is None:
            return None

        self._check_cutoff_in_rage(cutoff_date)
        last_past_cutoff_date = self._future_covariates.cutoff_date[
            self._future_covariates.cutoff_date <= cutoff_date
        ].max()

        future_covariates = self._future_covariates[
            (self._future_covariates.cutoff_date == last_past_cutoff_date)
            & (self._future_covariates.timestamp > cutoff_date)
        ]
        future_covariates.set_index("timestamp", inplace=True)
        return future_covariates



from enfobench.datasets.electricity_demand import ElectricityDemandDataset
from enfobench.datasets.gas_demand import GasDemandDataset
from enfobench.datasets.pv_generation import PVGenerationDataset

__all__ = [
    "ElectricityDemandDataset",
    "PVGenerationDataset",
    "GasDemandDataset",
]



from typing import Any

import duckdb
import pandas as pd

from enfobench.core import Subset
from enfobench.datasets.base import DatasetBase

Metadata = dict[str, Any]


class MetadataSubset(Subset):
    """Metadata subset of the HuggingFace dataset containing all metadata about the meters.

    Args:
        file_path: The path to the subset file.
    """

    def list_unique_ids(self) -> list[str]:
        """Lists all unique ids."""
        query = """
            SELECT DISTINCT unique_id
            FROM read_parquet(?)
        """
        conn = duckdb.connect(":memory:")
        return conn.execute(query, parameters=[str(self.file_path)]).fetch_df().unique_id.tolist()

    def get_by_unique_id(self, unique_id: str) -> Metadata:
        """Returns the metadata for the given unique id.

        Args:
            unique_id: The unique id of the meter.
        """
        query = """
            SELECT *
            FROM read_parquet(?)
            WHERE unique_id = ?
        """
        conn = duckdb.connect(":memory:")
        df = conn.execute(query, parameters=[str(self.file_path), unique_id]).fetch_df()
        if df.empty:
            msg = f"Unique id '{unique_id}' was not found."
            raise KeyError(msg)
        return df.iloc[0].to_dict()


class WeatherSubset(Subset):
    """Weather subset of the HuggingFace dataset containing all weather data.

    Args:
        file_path: The path to the subset file.
    """

    def list_location_ids(self) -> list[str]:
        """Lists all location ids."""
        query = """
            SELECT DISTINCT location_id
            FROM read_parquet(?)
        """
        conn = duckdb.connect(":memory:")
        return conn.execute(query, parameters=[str(self.file_path)]).fetch_df().location_id.tolist()

    def get_by_location_id(self, location_id: str, columns: list[str] | None = None) -> pd.DataFrame:
        """Returns the weather data for the given location id.

        Args:
            location_id: The location id of the weather station.
            columns: The columns to return. If None, all columns are returned.
        """
        conn = duckdb.connect(":memory:")

        if columns:
            query = f"""
                SELECT timestamp, {", ".join(columns)}
                FROM read_parquet(?)
                WHERE location_id = ?
            """  # noqa: S608
        else:
            query = """
                SELECT *
                FROM read_parquet(?)
                WHERE location_id = ?
            """
        df = conn.execute(query, parameters=[str(self.file_path), location_id]).fetch_df()
        if df.empty:
            msg = f"Location id '{location_id}' was not found."
            raise KeyError(msg)

        # Remove location_id and set timestamp as index
        df.drop(columns=["location_id"], inplace=True, errors="ignore")
        df.set_index("timestamp", inplace=True)
        return df


class DemandSubset(Subset):
    """Data subset of the HuggingFace dataset containing all electricity load data.

    Args:
        file_path: The path to the subset file.
    """

    def get_by_unique_id(self, unique_id: str):
        """Returns the demand data for the given unique id.

        Args:
            unique_id: The unique id of the meter.
        """
        query = """
            SELECT *
            FROM read_parquet(?)
            WHERE unique_id = ?
        """
        conn = duckdb.connect(":memory:")
        df = conn.execute(query, parameters=[str(self.file_path), unique_id]).fetch_df()
        if df.empty:
            msg = f"Unique id '{unique_id}' was not found."
            raise KeyError(msg)

        # Remove unique_id and set timestamp as index
        df.drop(columns=["unique_id"], inplace=True, errors="ignore")
        df.set_index("timestamp", inplace=True)
        return df


class ElectricityDemandDataset(DatasetBase):
    """ElectricityDemandDataset class representing the HuggingFace dataset.

    This class is a collection of all subsets inside HuggingFace dataset.
    It provides an easy way to access the different subsets.

    Args:
        directory: The directory where the HuggingFace dataset is located.
                   This directory should contain all the subset files.
    """

    HUGGINGFACE_DATASET = "EDS-lab/electricity-demand"
    SUBSETS = ("demand", "metadata", "weather")

    @property
    def metadata_subset(self) -> MetadataSubset:
        """Returns the metadata subset."""
        return MetadataSubset(self._get_subset_path("metadata"))

    @property
    def weather_subset(self) -> WeatherSubset:
        """Returns the weather subset."""
        return WeatherSubset(self._get_subset_path("weather"))

    @property
    def demand_subset(self) -> DemandSubset:
        """Returns the demand subset."""
        return DemandSubset(self._get_subset_path("demand"))

    def list_unique_ids(self) -> list[str]:
        return self.metadata_subset.list_unique_ids()

    def list_location_ids(self) -> list[str]:
        return self.weather_subset.list_location_ids()

    def get_data_by_unique_id(self, unique_id: str) -> tuple[pd.DataFrame, pd.DataFrame, Metadata]:
        metadata = self.metadata_subset.get_by_unique_id(unique_id)
        location_id = metadata["location_id"]

        demand = self.demand_subset.get_by_unique_id(unique_id)
        weather = self.weather_subset.get_by_location_id(location_id)
        # Filter weather data to match the period of the demand data
        weather = weather.loc[demand.index[0] - pd.Timedelta("7 days") : demand.index[-1] + pd.Timedelta("7 days")]
        return demand, weather, metadata



from pathlib import Path


class DatasetBase:
    SUBSETS = ()

    def __init__(self, directory: Path | str) -> None:
        directory = Path(directory).resolve()
        if not directory.is_dir() or not directory.exists():
            msg = "Please provide an existing directory where the dataset is located."
            raise ValueError(msg)
        self.directory = directory

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(directory={self.directory})"

    def _check_for_valid_subset(self, subset: str):
        if subset not in self.SUBSETS:
            msg = f"Please provide a valid subset. Available subsets: {self.SUBSETS}"
            raise ValueError(msg)

    def _get_subset_path(self, subset: str, extension: str = "parquet") -> Path:
        filepath = self.directory / f"{subset}.{extension}"
        if not filepath.exists():
            msg = f"Subset: {subset} is missing from the directory."
            raise ValueError(msg)
        return filepath



from datetime import timedelta

import pandas as pd
import pytest

from enfobench.evaluation import utils


@pytest.mark.parametrize(
    "horizon, freq, expected",
    [
        ("1 day", "15T", 96),
        ("1 day", "1H", 24),
        ("7 days", "1H", 7 * 24),
        ("1D", "1D", 1),
        ("1H", "1H", 1),
    ],
)
def test_steps_in_horizon(horizon, freq, expected):
    assert utils.steps_in_horizon(pd.Timedelta(horizon), freq) == expected


def test_steps_in_horizon_raises_with_non_multiple_horizon():
    with pytest.raises(ValueError):
        utils.steps_in_horizon(pd.Timedelta("36 minutes"), "15T")


@pytest.mark.parametrize(
    "freq, duration, expected",
    [
        ("15T", "15T", 1),
        ("15T", "1H", 4),
        ("15T", "1D", 96),
        ("15T", "2D", 2 * 96),
        ("15T", "1W", 7 * 96),
        ("15T", pd.Timedelta("15T"), 1),
        ("15T", pd.Timedelta("1H"), 4),
        ("15T", pd.Timedelta("1D"), 96),
        ("15T", pd.Timedelta("2D"), 2 * 96),
        ("15T", pd.Timedelta("1W"), 7 * 96),
        ("15T", timedelta(minutes=15), 1),
        ("15T", timedelta(hours=1), 4),
        ("15T", timedelta(days=1), 96),
        ("15T", timedelta(days=2), 2 * 96),
        ("15T", timedelta(weeks=1), 7 * 96),
        ("30T", "30T", 1),
        ("30T", "1H", 2),
        ("30T", "1D", 48),
        ("30T", "2D", 2 * 48),
        ("30T", "1W", 7 * 48),
        ("30T", pd.Timedelta("30T"), 1),
        ("30T", pd.Timedelta("1H"), 2),
        ("30T", pd.Timedelta("1D"), 48),
        ("30T", pd.Timedelta("2D"), 2 * 48),
        ("30T", pd.Timedelta("1W"), 7 * 48),
        ("30T", timedelta(minutes=30), 1),
        ("30T", timedelta(hours=1), 2),
        ("30T", timedelta(days=1), 48),
        ("30T", timedelta(days=2), 2 * 48),
        ("30T", timedelta(weeks=1), 7 * 48),
        ("1H", "1H", 1),
        ("1H", "1D", 24),
        ("1H", "2D", 2 * 24),
        ("1H", "1W", 7 * 24),
        ("1H", pd.Timedelta("1H"), 1),
        ("1H", pd.Timedelta("1D"), 24),
        ("1H", pd.Timedelta("2D"), 2 * 24),
        ("1H", pd.Timedelta("1W"), 7 * 24),
        ("1H", timedelta(hours=1), 1),
        ("1H", timedelta(days=1), 24),
        ("1H", timedelta(days=2), 2 * 24),
        ("1H", timedelta(weeks=1), 7 * 24),
    ],
)
def test_periods_in_duration(freq, duration, expected):
    target = pd.date_range(start="2020-01-01", end="2020-02-01", freq=freq)

    assert utils.periods_in_duration(target, duration) == expected


def test_periods_in_duration_raises_with_wrong_types():
    target = pd.date_range(start="2020-01-01", end="2020-02-01", freq="15T")
    with pytest.raises(ValueError):
        utils.periods_in_duration(target, 3)


def test_periods_in_duration_raises_if_duration_is_not_multiples():
    target = pd.date_range(start="2020-01-01", end="2020-02-01", freq="15T")
    with pytest.raises(ValueError):
        utils.periods_in_duration(target, "4T")


@pytest.mark.parametrize(
    "start_date,end_date,horizon,step,expected",
    [
        ("2020-01-01", "2020-01-05", "2 days", "1 day", ["2020-01-01", "2020-01-02", "2020-01-03"]),
        ("2020-01-01", "2020-01-05", "1 day", "1 day", ["2020-01-01", "2020-01-02", "2020-01-03", "2020-01-04"]),
    ],
)
def test_generate_cutoff_dates(start_date, end_date, horizon, step, expected):
    start_date = pd.Timestamp(start_date)
    end_date = pd.Timestamp(end_date)
    horizon = pd.Timedelta(horizon)
    step = pd.Timedelta(step)

    cutoff_dates = utils.generate_cutoff_dates(start_date, end_date, horizon, step)

    assert cutoff_dates == [pd.Timestamp(date) for date in expected]


def test_generate_cutoff_date_raises_with_longer_horizon():
    with pytest.raises(ValueError):
        utils.generate_cutoff_dates(
            start_date=pd.Timestamp("2020-01-01"),
            end_date=pd.Timestamp("2020-01-05"),
            horizon=pd.Timedelta("7 days"),
            step=pd.Timedelta("1 day"),
        )


def test_generate_cutoff_date_raises_with_wrong_end_and_start_dates():
    with pytest.raises(ValueError):
        utils.generate_cutoff_dates(
            start_date=pd.Timestamp("2020-01-01"),
            end_date=pd.Timestamp("2020-01-01"),
            horizon=pd.Timedelta("1 day"),
            step=pd.Timedelta("1 day"),
        )


def test_generate_cutoff_date_raises_with_negative_horizon():
    with pytest.raises(ValueError):
        utils.generate_cutoff_dates(
            start_date=pd.Timestamp("2020-01-01"),
            end_date=pd.Timestamp("2020-01-05"),
            horizon=pd.Timedelta("-1 day"),
            step=pd.Timedelta("1 day"),
        )


def test_generate_cutoff_date_raises_with_negative_step():
    with pytest.raises(ValueError):
        utils.generate_cutoff_dates(
            start_date=pd.Timestamp("2020-01-01"),
            end_date=pd.Timestamp("2020-01-05"),
            horizon=pd.Timedelta("1 day"),
            step=pd.Timedelta("-1 day"),
        )


@pytest.mark.parametrize(
    "freq,horizon",
    [
        ("1H", 24),
        ("1H", 38),
        ("30T", 48),
        ("30T", 72),
        ("15T", 96),
        ("15T", 144),
    ],
)
def test_create_forecast_index(helpers, freq, horizon):
    dataset = helpers.generate_univariate_dataset(start="2020-01-01", end="2020-01-31", freq=freq)

    cutoff_date = helpers.random_date(dataset.target_available_since, dataset.target_available_until, freq)
    history = dataset.get_history(cutoff_date)

    index = utils.create_forecast_index(history, horizon)

    assert isinstance(index, pd.DatetimeIndex)
    assert index.freq == history.index.freq
    assert len(index) == horizon
    assert index[0] == cutoff_date
    assert (index >= cutoff_date).all()



import pandas as pd

from enfobench.evaluation.filters import hampel


def test_hampel():
    s = pd.Series([1, 2, 3, 4, 5, 6, 7, 8, 9, 1000, 11, 12, 13, 14, 15, 16, 17, 18, 19])
    expected_s = pd.Series([1, 2, 3, 4, 5, 6, 7, 8, 9, 11, 11, 12, 13, 14, 15, 16, 17, 18, 19])
    expected_mask = pd.Series([False] * 9 + [True] + [False] * 9)

    filtered, mask = hampel(s, window_size=5, n_sigmas=3.0)

    assert (filtered == expected_s).all()
    assert (mask == expected_mask).all()



import pandas as pd
import pytest

from enfobench.core import Dataset


@pytest.mark.parametrize("resolution", ["15T", "30T", "1H"])
def test_get_history(helpers, resolution):
    target = helpers.generate_target(start="2020-01-01", end="2021-01-01", freq=resolution)
    ds = Dataset(target)

    cutoff_date = helpers.random_date(ds.target_available_since, ds.target_available_until, resolution)

    history = ds.get_history(cutoff_date)

    assert history.index.name == "timestamp"
    assert isinstance(history.index, pd.DatetimeIndex)
    assert "y" in history.columns
    assert len(history.columns) == 1
    assert cutoff_date not in history.index
    assert (history.index < cutoff_date).all()


@pytest.mark.parametrize("resolution", ["15T", "30T", "1H"])
def test_get_past_covariates(helpers, resolution):
    target = helpers.generate_target(start="2020-01-01", end="2021-01-01", freq=resolution)
    past_covariates = helpers.generate_covariates(
        start="2020-01-01",
        end="2021-01-01",
        freq=resolution,
        columns=["a", "b"],
    )
    ds = Dataset(target, past_covariates)

    cutoff_date = helpers.random_date(ds.target_available_since, ds.target_available_until, resolution)

    past_cov = ds.get_past_covariates(cutoff_date)

    assert past_cov.index.name == "timestamp"
    assert isinstance(past_cov.index, pd.DatetimeIndex)
    assert "a" in past_cov.columns
    assert "b" in past_cov.columns
    assert cutoff_date in past_cov.index
    assert (past_cov.index <= cutoff_date).all()


@pytest.mark.parametrize("resolution", ["6H", "1D"])
def test_get_future_covariates(helpers, resolution):
    target = helpers.generate_target(start="2020-01-01", end="2021-01-01", freq="30T")
    future_covariates = helpers.generate_future_covariates(
        start="2020-01-01",
        end="2021-01-01",
        freq=resolution,
        columns=["a", "b"],
    )
    ds = Dataset(target, future_covariates=future_covariates)

    cutoff_date = helpers.random_date(ds.target_available_since, ds.target_available_until, resolution)

    future_cov = ds.get_future_covariates(cutoff_date)

    assert future_cov.index.name == "timestamp"
    assert isinstance(future_cov.index, pd.DatetimeIndex)
    assert "a" in future_cov.columns
    assert "b" in future_cov.columns
    assert (future_cov.cutoff_date <= cutoff_date).all()
    assert (future_cov.index > cutoff_date).all()


if __name__ == "__main__":
    pytest.main()



import pandas as pd
import pytest
from starlette.testclient import TestClient

from enfobench.evaluation import (
    ForecastClient,
    cross_validate,
)
from enfobench.evaluation.evaluate import _compute_metric, _compute_metrics, evaluate_metrics
from enfobench.evaluation.metrics import mean_absolute_error, mean_squared_error
from enfobench.evaluation.server import server_factory


def test_cross_validate_univariate_locally(helpers, model):
    dataset = helpers.generate_univariate_dataset(start="2020-01-01", end="2021-01-01", freq="30T")

    start_date = pd.Timestamp("2020-02-01 10:00:00")
    end_date = pd.Timestamp("2020-04-01")
    horizon = pd.Timedelta("38 hours")
    step = pd.Timedelta("1 day")

    forecasts = cross_validate(model, dataset, start_date=start_date, end_date=end_date, horizon=horizon, step=step)

    assert isinstance(forecasts, pd.DataFrame)
    assert "y" in forecasts.columns
    assert "yhat" in forecasts.columns

    assert "timestamp" in forecasts.columns
    assert (forecasts.timestamp >= start_date).all()
    assert forecasts.timestamp.iloc[0] == start_date

    assert "cutoff_date" in forecasts.columns
    assert (start_date.time() == forecasts.cutoff_date.dt.time).all()

    for cutoff_date, forecast in forecasts.groupby("cutoff_date"):
        assert len(forecast) == 38 * 2  # 38 hours with half-hour series
        assert forecast.timestamp.iloc[0] == cutoff_date
        assert (forecast.timestamp >= cutoff_date).all()

    assert list(forecasts.cutoff_date.unique()) == list(
        pd.date_range(start="2020-02-01 10:00:00", end="2020-03-30 10:00:00", freq="1D", inclusive="both")
    )


def test_cross_validate_univariate_via_server(helpers, model):
    dataset = helpers.generate_univariate_dataset(start="2020-01-01", end="2020-03-01", freq="30T")

    app = server_factory(model=model)
    test_client = TestClient(app)
    model = ForecastClient(client=test_client)

    forecasts = cross_validate(
        model=model,
        dataset=dataset,
        start_date=pd.Timestamp("2020-02-01 10:00:00"),
        end_date=pd.Timestamp("2020-03-01"),
        horizon=pd.Timedelta("38 hours"),
        step=pd.Timedelta("1 day"),
    )

    assert isinstance(forecasts, pd.DataFrame)
    assert "timestamp" in forecasts.columns
    assert "y" in forecasts.columns
    assert "yhat" in forecasts.columns
    assert "cutoff_date" in forecasts.columns
    assert list(forecasts.cutoff_date.unique()) == list(
        pd.date_range(start="2020-02-01 10:00:00", end="2020-02-28 10:00:00", freq="1D", inclusive="both")
    )


def test_cross_validate_multivariate_locally(helpers, model):
    dataset = helpers.generate_multivariate_dataset(
        columns=["x", "z"], start="2020-01-01", end="2020-03-01", freq="30T"
    )

    forecasts = cross_validate(
        model=model,
        dataset=dataset,
        start_date=pd.Timestamp("2020-02-01 10:00:00"),
        end_date=pd.Timestamp("2020-03-01"),
        horizon=pd.Timedelta("38 hours"),
        step=pd.Timedelta("1 day"),
    )

    assert isinstance(forecasts, pd.DataFrame)
    assert "timestamp" in forecasts.columns
    assert "y" in forecasts.columns
    assert "yhat" in forecasts.columns
    assert "cutoff_date" in forecasts.columns
    assert list(forecasts.cutoff_date.unique()) == list(
        pd.date_range(start="2020-02-01 10:00:00", end="2020-02-28 10:00:00", freq="1D", inclusive="both")
    )


def test_cross_validate_multivariate_via_server(helpers, model):
    dataset = helpers.generate_multivariate_dataset(
        columns=["x", "z"], start="2020-01-01", end="2020-03-01", freq="30T"
    )

    app = server_factory(model=model)
    test_client = TestClient(app)
    model = ForecastClient(client=test_client)

    forecasts = cross_validate(
        model=model,
        dataset=dataset,
        start_date=pd.Timestamp("2020-02-01 10:00:00"),
        end_date=pd.Timestamp("2020-03-01"),
        horizon=pd.Timedelta("38 hours"),
        step=pd.Timedelta("1 day"),
    )

    assert isinstance(forecasts, pd.DataFrame)
    assert "timestamp" in forecasts.columns
    assert "y" in forecasts.columns
    assert "yhat" in forecasts.columns
    assert "cutoff_date" in forecasts.columns
    assert list(forecasts.cutoff_date.unique()) == list(
        pd.date_range(start="2020-02-01 10:00:00", end="2020-02-28 10:00:00", freq="1D", inclusive="both")
    )


def test_compute_metric(clean_forecasts):
    metric_value = _compute_metric(clean_forecasts, mean_absolute_error)

    assert 0 < metric_value < 1


def test_compute_metrics(clean_forecasts):
    metric_values = _compute_metrics(clean_forecasts, {"MAE": mean_absolute_error, "MSE": mean_squared_error})

    assert isinstance(metric_values, dict)
    assert 0 < metric_values["MAE"] < 1
    assert 0 < metric_values["MSE"] < 1


def test_compute_metric_on_forecast_with_missing_values_raises_error(forecasts_with_missing_values):
    with pytest.raises(ValueError):
        _compute_metric(forecasts_with_missing_values, mean_absolute_error)


def test_compute_metrics_on_forecast_with_missing_values_raises_error(forecasts_with_missing_values):
    with pytest.raises(ValueError):
        _compute_metrics(forecasts_with_missing_values, {"MAE": mean_absolute_error, "MSE": mean_squared_error})


def test_evaluate_metrics_on_clean_forecasts(clean_forecasts):
    metrics = evaluate_metrics(clean_forecasts, {"MAE": mean_absolute_error, "MSE": mean_squared_error})

    assert isinstance(metrics, pd.DataFrame)
    assert "MAE" in metrics.columns
    assert "MSE" in metrics.columns
    assert "weight" in metrics.columns
    assert len(metrics) == 1
    assert 0 < metrics["MAE"].iloc[0] < 1
    assert 0 < metrics["MSE"].iloc[0] < 1
    assert metrics["weight"].iloc[0] == 1


def test_evaluate_metrics_on_forecasts_with_missing_values(forecasts_with_missing_values):
    metrics = evaluate_metrics(forecasts_with_missing_values, {"MAE": mean_absolute_error, "MSE": mean_squared_error})

    assert isinstance(metrics, pd.DataFrame)
    assert "MAE" in metrics.columns
    assert "MSE" in metrics.columns
    assert "weight" in metrics.columns
    assert len(metrics) == 1
    assert 0 < metrics["MAE"].iloc[0] < 1
    assert 0 < metrics["MSE"].iloc[0] < 1
    assert pytest.approx(metrics["weight"].iloc[0], 0.1) == 1 - 0.3


def test_evaluate_metrics_on_clean_forecasts_grouped_by(clean_forecasts):
    metrics = evaluate_metrics(
        clean_forecasts,
        {"MAE": mean_absolute_error, "MSE": mean_squared_error},
        groupby="cutoff_date",
    )

    grouped_values = clean_forecasts["cutoff_date"].unique()

    assert isinstance(metrics, pd.DataFrame)
    assert "MAE" in metrics.columns
    assert "MSE" in metrics.columns
    assert "weight" in metrics.columns
    assert "cutoff_date" in metrics.columns
    assert len(metrics) == len(grouped_values)
    assert all(0 < metric < 1 for metric in metrics["MAE"])
    assert all(0 < metric < 1 for metric in metrics["MSE"])
    assert all(metric == 1 for metric in metrics["weight"])


def test_evaluate_metrics_on_forecasts_with_missing_values_grouped_by(forecasts_with_missing_values):
    metrics = evaluate_metrics(
        forecasts_with_missing_values,
        {"MAE": mean_absolute_error, "MSE": mean_squared_error},
        groupby="cutoff_date",
    )

    grouped_values = forecasts_with_missing_values["cutoff_date"].unique()

    assert isinstance(metrics, pd.DataFrame)
    assert "MAE" in metrics.columns
    assert "MSE" in metrics.columns
    assert "weight" in metrics.columns
    assert "cutoff_date" in metrics.columns
    assert len(metrics) == len(grouped_values)
    assert all(0 < metric < 1 for metric in metrics["MAE"])
    assert all(0 < metric < 1 for metric in metrics["MSE"])
    assert all(0 <= metric <= 1 for metric in metrics["weight"])


if __name__ == "__main__":
    pytest.main()
